/* reconhec.c - Reconhecimento de sentencas */

#include <stdio.h>
#include <math.h>
#include "reconhec.h"
#include "bcopalav.h"
#include "estrdado.h"
#include "manutarq.h"
#include "alocacao.h"


char palavras [20] [TAM_PAL]; /* Palavras da frase inicial. */

tipo_classe classes [20]; /* Classes de cada palavra da frase inicial. */

int num_pal; /* Nmero de palavras da frase inicial. */

nivel_arvore arvore = NULL, /* Aponta para o inicio da arvore, i, aponta
			       para o nvel zero das listas. */

	nivel_atual = NULL, /* Aponta para o nivel atual da arvore, ou seja
			       para a proimeira frase que pertence
			       a este nivel. */

     nivel_anterior = NULL, /* Aponta para o nivel anterior. */

     sentencas = NULL;      /* Aponta para as sentencas geradas */

regras_sintaticas regra [NUMERO_REGRAS]; /* Tabela de regras. Elas so
                                        montadas pela funo monta_regras. */


/*--------------------------------------------------------------*/
/* Faz a inicializao da tabela de regras.                      */
/*--------------------------------------------------------------*/
void monta_regras(void)
{

  /* f0 */
  regra[0].prim_oper = QU;
  regra[0].seg_oper = NC;
  regra[0].result = SN;

  /* f4 */
  regra[1].prim_oper = SN;
  regra[1].seg_oper = VI;
  regra[1].result = SEN;

  /* f5a */
  regra[2].prim_oper = VT;
  regra[2].seg_oper = SN;
  regra[2].result = VI;

  /* f5b */
  regra[3].prim_oper = VC;
  regra[3].seg_oper = ADJ;
  regra[3].result = VI;

  /* f6a */
  regra[4].prim_oper = ADJ;
  regra[4].seg_oper = NC;
  regra[4].result = NC;

  /* f6b */
  regra[5].prim_oper = NC;
  regra[5].seg_oper = ADJ;
  regra[5].result = NC;

  /* f7 */
  regra[6].prim_oper = VI;
  regra[6].seg_oper = ADJ;
  regra[6].result = VI;

  /* f8b */
  regra[7].prim_oper = SN;
  regra[7].seg_oper = SN;
  regra[7].result = SN;

  /* f8c */
  regra[8].prim_oper = VI;
  regra[8].seg_oper = VI;
  regra[8].result = VI;

  /* f11 */
  regra[9].prim_oper = PADV;
  regra[9].seg_oper = SN;
  regra[9].result = ADJ;

  /* f12 */
  regra[10].prim_oper = PSN;
  regra[10].seg_oper = SN;
  regra[10].result = ADJ;

  /* f13 */
  regra[11].prim_oper = PNC;
  regra[11].seg_oper = NC;
  regra[11].result = ADJ;
}

/*--------------------------------------------------------------*/
/* Faz uma inclusao na lista de sentencas.                      */
/*--------------------------------------------------------------*/
void anota_nova_sentenca(categ_sintatica s)
{

  nivel_arvore novo_nodo = aloca_nivel ();

  novo_nodo->frase = s;
  novo_nodo->proximo = sentencas->proximo;
  sentencas->proximo = novo_nodo;
}

/*--------------------------------------------------------------*/
/* Se existe uma regra aplicvel a 'prim' e 'seg', ento retorna*/
/* o nodo 'result', aplicada a regra. Caso contrrio retorna    */
/* NULL.                                                        */
/*--------------------------------------------------------------*/
categ_sintatica ha_regra(categ_sintatica prim, categ_sintatica seg)
{
  int i;
  categ_sintatica result = NULL;

  for(i=0; i < NUMERO_REGRAS; i++)
    if(regra[i].prim_oper == prim->tipo && regra[i].seg_oper == seg->tipo)
    {
      result = aloca_categ_sintatica();
      result->codigo_prim = prim->codigo_prim + prim->codigo_seg;
      result->codigo_seg = seg->codigo_prim + seg->codigo_seg;
      result->prim_categ = prim;
      result->seg_categ = seg;
      result->tipo = regra[i].result;
      if(regra[i].result == SEN) /* se foi reconhecida uma sentenca, anota. */
	anota_nova_sentenca(result);
      break;
    }

  return result;
}


/*--------------------------------------------------------------*/
/* Constroe a partir das palavras da entrada, a frase inicial,   */
/* que constitui o nvel zero da arvore. Associa a cada palavra */
/* sua categoria sinttica, que est armazenada em 'classes'.   */
/*--------------------------------------------------------------*/
nivel_arvore gera_frase_inicial(void)
{
  int i;
  categ_sintatica aux;

  arvore = aloca_nivel();
  nivel_anterior = nivel_atual = arvore;
  arvore->frase = aloca_categ_sintatica();
  arvore->frase->tipo = classes[0];
  arvore->frase->codigo_prim = 1;
  arvore->frase->codigo_seg = 0;

  for(i=1, aux=arvore->frase; i < num_pal; i++)
  {
    aux->proximo = aloca_categ_sintatica();
    aux = aux->proximo;
    aux->codigo_prim = pow (2,i);
    aux->codigo_seg = 0;
    aux->tipo = classes[i];
  }

  return arvore;
}



/*--------------------------------------------------------------*/
/* Adiciona  frase um novo nodo.                               */
/*--------------------------------------------------------------*/
categ_sintatica copia(categ_sintatica dest, categ_sintatica fonte)
{
  categ_sintatica aux;


  if(dest == NULL)
  {
    dest = aloca_categ_sintatica();
    memcpy(dest, fonte, sizeof(*fonte));
    dest->proximo = NULL;
  }
  else
  {
    for(aux = dest; aux->proximo != NULL; aux = aux->proximo)
      ; /* percorre a lista at o final. */
    aux->proximo = aloca_categ_sintatica();
    aux = aux->proximo;
    memcpy(aux, fonte, sizeof(*fonte));
    aux->proximo = NULL;
  }

  return dest;
}

/*--------------------------------------------------------------*/
/* Cria uma nova frase, onde 'result'  substitui dois nodos da  */
/* frase original, mediante a aplicao de alguma regra.        */
/*--------------------------------------------------------------*/
categ_sintatica cria_nova_frase(categ_sintatica frase, categ_sintatica result,
                                categ_sintatica aux1)
{
  categ_sintatica nova_frase = NULL, aux;


  for(aux = frase; aux != aux1; aux = aux->proximo)
    nova_frase = copia(nova_frase, aux);
  nova_frase = copia(nova_frase, result);
  aux1 = aux1->proximo->proximo; /* pula os 2 nodos que foram unidos pela regra. */
  for(aux = aux1; aux != NULL; aux = aux->proximo)
    nova_frase = copia(nova_frase, aux);

  return nova_frase;

}


nivel_arvore adiciona(nivel_arvore nivel, categ_sintatica nodo)
{
  categ_sintatica novo = aloca_categ_sintatica (), aux;

  memcpy(novo, nodo, sizeof(*nodo));
  novo->proximo = NULL;
  for(aux = nivel->frase; (aux != NULL) && (aux->proximo != NULL);
      aux = aux->proximo)
   ;
  aux->proximo = novo;

  return nivel;
}


/* unifica duas frases, retorna a frase unificada ou NULL. */
nivel_arvore unifica_frase (nivel_arvore frase1, nivel_arvore frase2)
{
  categ_sintatica aux_frase1, aux_frase2;
  nivel_arvore frase_unificada;
  int identica = 1;



  for(aux_frase1 = frase1->frase, aux_frase2 = frase2->frase;
      (aux_frase1 != NULL) && (aux_frase2 != NULL); )
  {
    if((aux_frase1->codigo_prim == aux_frase2->codigo_prim)
       && (aux_frase1->codigo_seg == aux_frase2->codigo_seg))
    {
      aux_frase1 = aux_frase1->proximo;
      aux_frase2 = aux_frase2->proximo;
    }
    else if((aux_frase1->codigo_prim ==
	     aux_frase2->codigo_prim + aux_frase2->codigo_seg)
	  && (aux_frase1->codigo_seg ==
	      aux_frase2->proximo->codigo_prim + aux_frase2->proximo->codigo_seg))
    {
      aux_frase1 = aux_frase1->proximo;
      aux_frase2 = aux_frase2->proximo->proximo;
      identica = 0;
    }
    else if((aux_frase2->codigo_prim ==
	     aux_frase1->codigo_prim + aux_frase1->codigo_seg)
	  && (aux_frase2->codigo_seg ==
	      aux_frase1->proximo->codigo_prim + aux_frase1->proximo->codigo_seg))
    {
      aux_frase1 = aux_frase1->proximo->proximo;
      aux_frase2 = aux_frase2->proximo;
      identica = 0;
    }
    else /* impossivel unificar */
      return NULL;
  }
  if(identica)
    return NULL;

  /* se chegar aqui, entao pode unificar */

  frase_unificada = aloca_nivel ();

  for(aux_frase1 = frase1->frase, aux_frase2 = frase2->frase;
      (aux_frase1 != NULL) && (aux_frase2 != NULL); )
  {
    if((aux_frase1->codigo_prim == aux_frase2->codigo_prim)
       && (aux_frase1->codigo_seg == aux_frase2->codigo_seg))
    {
      frase_unificada = adiciona(frase_unificada, aux_frase1);
      aux_frase1 = aux_frase1->proximo;
      aux_frase2 = aux_frase2->proximo;
    }
    else if((aux_frase1->codigo_prim ==
	     aux_frase2->codigo_prim + aux_frase2->codigo_seg)
	  && (aux_frase1->codigo_seg ==
	      aux_frase2->proximo->codigo_prim + aux_frase2->proximo->codigo_seg))
    {
      frase_unificada = adiciona(frase_unificada, aux_frase1);
      aux_frase1 = aux_frase1->proximo;
      aux_frase2 = aux_frase2->proximo->proximo;
    }
    else if((aux_frase2->codigo_prim ==
	     aux_frase1->codigo_prim + aux_frase1->codigo_seg)
	  && (aux_frase2->codigo_seg ==
	      aux_frase1->proximo->codigo_prim + aux_frase1->proximo->codigo_seg))
    {
      frase_unificada = adiciona(frase_unificada, aux_frase2);
      aux_frase1 = aux_frase1->proximo->proximo;
      aux_frase2 = aux_frase2->proximo;
    }
  }

  return frase_unificada;

}



/* Controla a sequencia de unificacao das frases em um nivel. */
nivel_arvore unifica(nivel_arvore nivel_atual, nivel_arvore atual)
{

  nivel_arvore aux_frase, nova_frase = NULL;

  for (aux_frase = nivel_atual->proximo;
       aux_frase != atual; aux_frase = aux_frase->proximo)
  {
    if ((nova_frase = unifica_frase (aux_frase, atual)) != NULL)
    {
      nivel_arvore aux;

      for (aux = atual; aux->proximo != NULL; aux = aux->proximo)
	;

      aux->proximo = nova_frase;
    }

  }
  return nova_frase;
}

/*--------------------------------------------------------------*/
/* Cria todos os sucessore para uma dada frase, fazendo a unifi-*/
/* cacao.                                                       */
/*--------------------------------------------------------------*/
nivel_arvore gera_sucessores_frase(nivel_arvore atual, nivel_arvore anterior)
{

  categ_sintatica aux1, aux2, res1,
		   nova_frase;


  for(aux1 = anterior->frase; aux1->proximo != NULL; aux1 = aux1->proximo)
  {
    aux2 = aux1->proximo;
    if((res1 = ha_regra(aux1,aux2)) != NULL)
    {
      nivel_arvore frase_unificada;
      nova_frase = cria_nova_frase(anterior->frase, res1, aux1);
      atual->proximo = aloca_nivel();
      atual = atual->proximo;
      atual->frase = nova_frase;
      if(nivel_atual->proximo != atual)
	if((frase_unificada = unifica(nivel_atual, atual)) != NULL)
          atual = frase_unificada;
    }
  }

  return atual;
}

/*--------------------------------------------------------------*/
/* Cria todos os sucessores para um dado nivel da arvore.       */
/*--------------------------------------------------------------*/
nivel_arvore gera_nivel(void)
{
/*
dados:
  nivel_anterior: inicio das frases de nivel n-1.
  nivel_atual: inicio das frases do nivel n, que esta sendo criado
               agora, aponta para a ultima frase.
  atual: posicao da ultima frase gerada ( usado para unificacao).
  anterior : frase corrente no nivel anterior.
*/

  nivel_arvore atual,
	       anterior;

  if(nivel_anterior == nivel_atual)
    atual = gera_sucessores_frase(nivel_atual, nivel_anterior);
  else
    for(anterior = nivel_anterior, atual = nivel_atual;
        anterior != nivel_atual->proximo; anterior = anterior->proximo)
      atual = gera_sucessores_frase(atual, anterior);

  nivel_atual = nivel_atual->proximo;
  nivel_anterior = nivel_atual;
  nivel_atual = atual;

  return nivel_anterior;
}

/* mostra lista */

void mostra_lista (nivel_arvore inicio)
{
  nivel_arvore nodo_atual;
  categ_sintatica categ;

  putchar ('\n');
  for (nodo_atual = inicio; nodo_atual != NULL;
       nodo_atual = nodo_atual->proximo)
  {
    for (categ = nodo_atual->frase; categ != NULL;
	 categ = categ->proximo)
    {

      switch (categ->tipo)
      {
	case SEN:
	  printf ("SEN ");
	  break;
	case ADJ:
	  printf ("ADJ ");
	  break;
	case ADV:
	  printf ("ADV ");
	  break;
	case QU:
	  printf ("QU ");
	  break;
	case SN:
	  printf ("SN ");
	  break;
	case NC:
	  printf ("NC ");
	  break;
	case PADV:
	  printf ("PADV ");
	  break;
	case PNC:
	  printf ("PNC ");
	  break;
	case PSN:
	  printf ("PSN ");
	  break;
	case VC:
	  printf ("VC ");
	  break;
	case VI:
	  printf ("VI ");
	  break;
	case VT:
	  printf ("VT ");
	  break;
      }

    }
    putchar ('\n');
  }


  getchar ();



}


/* para todos os niveis */
void gera_todos_os_niveis(void)
{

  arvore = nivel_atual = nivel_anterior = sentencas = NULL;

  sentencas = aloca_nivel ();

  nivel_atual = gera_frase_inicial ();
  monta_regras ();

  while(/*pode_continuar()*/1)
  {
    if(gera_nivel() == NULL) /* caso nao seja gerada nenhuma nova frase */
      break;
    mostra_lista (arvore);
  }



  desaloca_lista (arvore);
  desaloca_grupo_sen (sentencas);

}

